/*
 * Hydrogen
 * Copyright(c) 2002-2003 by Alex >Comix< Cominu [comix@users.sourceforge.net]
 *
 * http://hydrogen.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY, without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: Hydrogen.cpp,v 1.111 2003/09/26 09:24:22 comix Exp $
 *
 */
#include "config.h"

#include <pthread.h>
#include <deque>
#include <sys/time.h>
#include <iostream>
#include <time.h>
using std::cout;
using std::cerr;
using std::endl;

#include "Sample.h"
#include "LocalFileMng.h"
#include "Hydrogen.h"

#include "JackDriver.h"
#include "OssDriver.h"
#include "NullDriver.h"
#include "DiskWriterDriver.h"
#include "AlsaMidiDriver.h"


// GLOBALS
/** Mutex for syncronizated access to song object */
pthread_mutex_t engineLock_mutex;


/** Hydrogen class instance (used for log) */
Hydrogen* hydrogenInstance = NULL;


/** Audio Driver */
GenericDriver *driver = NULL;

/** Song Note FIFO */
std::deque<Note*> songNoteQueue;


// check if ALSA_SEQ support is enabled
#ifdef USE_ALSA_SEQ
/** Midi Driver */
AlsaMidiDriver *midiDriver = NULL;
#endif

unsigned maxNotes;

/** Midi Note FIFO */
std::deque<Note*> midiNoteQueue;

/** Playing note FIFO */
std::vector<Note*> playingNotesQueue;

unsigned long totalFrames = 0;

uint tickSize;

/** Current song */
Song *song;

/** Current pattern */
Pattern* pattern;

/** Next pattern (used only in Live mode) */
Pattern* nextPattern;

int patternPos;

/** Metronome Instrument */
Instrument *metronomeInstrument = NULL;

/** Master peak (left channel) */
float masterPeak_L = 0.0f;

/** Master peak (right channel) */
float masterPeak_R = 0.0f;

// questi buffer vengono usati nella funzione process
uint audioEngine_buffer_size = 0;
float *main_buffer_L = NULL;
float *main_buffer_R = NULL;
float *track_buffers_L[MAX_TRACKS];
float *track_buffers_R[MAX_TRACKS];
bool use_track_outs = false;

/** static reference of Hydrogen class */
Hydrogen* Hydrogen::instance = NULL;

/** number of playing notes */
uint playingNotes = 0;

/** Audio engine state */
int audioEngine_state = UNINITIALIZED;

std::vector<EngineListener*> audioEngine_listeners;

/** time used in process function */
float processTime = 0.0f;

/** max ms usable in process with no xrun */
float maxProcessTime;

/** Delay FX */
#define MAX_DELAY  44100 * 2
float delayRingBuffer_L[MAX_DELAY];	// 1sec delay
float delayRingBuffer_R[MAX_DELAY];	// 1sec delay

float *delayBuffer_L = NULL;
float *delayBuffer_R = NULL;
float delayBufferPeak_L = 0.0f;
float delayBufferPeak_R = 0.0f;

int delayStartPos = 0;

int patternStartTick = -1;
uint m_nPatternTickPosition = 0;

// used in findPatternInTick
int findPatternInTick_lastTick = 0;
int findPatternInTick_lastTotal = 0;



// PROTOTYPES
void	audioEngine_init();
void	audioEngine_destroy();
int	audioEngine_start(bool bLockEngine = false);
void	audioEngine_stop(bool bLockEngine = false);
void	audioEngine_setSong(Song *newSong);
void	audioEngine_removeSong();
void	audioEngine_noteOn(Note *note);
void	audioEngine_noteOff(Note *note);
int	audioEngine_process(uint32_t nframes, void *arg);
inline void audioEngine_clearNoteQueue();
inline void audioEngine_process_checkBPMChanged();
inline void audioEngine_process_playPattern(unsigned long nframes);
inline void audioEngine_process_playMidi(unsigned long nframes);
inline void audioEngine_process_updateInfo();

inline uint audioEngine_renderNote(Note* note, int nFrames, uint bufferSize);
inline int audioEngine_updateNoteQueue(uint nFrames);

inline int findPatternInTick( int tick, bool loopMode );

void audioEngine_seek( unsigned long nFrames );

void audioEngine_restartAudioDrivers();
void audioEngine_startAudioDrivers();
void audioEngine_stopAudioDrivers();


//----------------------------------------------------------------------------
inline unsigned long currentTime() {
	struct timeval now;
	gettimeofday(&now, NULL);
	return now.tv_sec * 1000 + now.tv_usec / 1000;
}



//----------------------------------------------------------------------------
inline timeval currentTime2() {
	struct timeval now;
	gettimeofday(&now, NULL);
	return now;
}



inline int randomValue( int max ) {
	return rand() % max;
}


inline void audioEngine_lock() {
	pthread_mutex_lock(&engineLock_mutex);
}


inline void audioEngine_unlock() {
	pthread_mutex_unlock(&engineLock_mutex);
}


inline void audioEngine_changeStatus( int newStatus ) {
	audioEngine_lock();
	audioEngine_state = newStatus;
	audioEngine_unlock();
}



void audioEngine_raiseError( unsigned nErrorCode ) {
	#warning "Convertire il codice di errore in stringa"
//	hydrogenInstance->errorLog( msg );
	// raise error event
	for (uint i = 0; i < audioEngine_listeners.size(); i++) {
		audioEngine_listeners[i]->engineError( nErrorCode );
	}
}

void updateTickSize() {
	hydrogenInstance->warningLog("UpdateTickSize");
	float bpm = song->getBpm();
	float resolution = song->getResolution();
	float sampleRate = (float)driver->getSampleRate();
	tickSize = (uint)( sampleRate * 60.0 /  bpm / resolution );
}


//----------------------------------------------------------------------------
/**
 *
 */
void audioEngine_init() {
	hydrogenInstance->infoLog("Hydrogen audio engine init");

	// check current state
	if (audioEngine_state != UNINITIALIZED) {
		hydrogenInstance->errorLog("audioEngine_init: Error the audio engine is not in UNINITIALIZED state");
		audioEngine_unlock();
		return;
	}

	song = NULL;
	pattern = NULL;
	nextPattern = NULL;
	patternPos = -1;
	m_nPatternTickPosition = 0;
	metronomeInstrument = NULL;
	driver = NULL;

	main_buffer_L = NULL;
	main_buffer_R = NULL;
	delayBuffer_L = NULL;
	delayBuffer_R = NULL;

	for (unsigned int n=0; n < MAX_TRACKS; ++n) {
		track_buffers_L[n] = NULL;
		track_buffers_R[n] = NULL;
	}

	memset( delayRingBuffer_L, 0, MAX_DELAY * sizeof(float));
	memset( delayRingBuffer_R, 0, MAX_DELAY * sizeof(float));


	maxNotes = PreferencesMng::getInstance()->getMaxNotes();
	use_track_outs = PreferencesMng::getInstance()->isJackTrackOuts();


	pthread_mutex_init(&engineLock_mutex, NULL);

	srand(time(NULL));

	// Create metronome instrument
	metronomeInstrument = Instrument::load( string(DATA_DIR) + "/click.wav" );
	metronomeInstrument->setVolume( 0.0 );


	// Change the current audio engine state
	audioEngine_state = INITIALIZED;
	hydrogenInstance->infoLog( "INITIALIZED" );

	// raise stateChanged event
	for (uint i = 0; i < audioEngine_listeners.size(); i++) {
		audioEngine_listeners[i]->stateChanged( INITIALIZED );
	}
}





//----------------------------------------------------------------------------
/**
 *
 */
void audioEngine_destroy() {
	audioEngine_lock();
	hydrogenInstance->infoLog("Hydrogen audio engine destroy");

	// check current state
	if (audioEngine_state != INITIALIZED) {
		hydrogenInstance->errorLog("audioEngine_destroy: Error the audio engine is not in INITIALIZED state");
		return;
	}

	// change the current audio engine state
	audioEngine_state = UNINITIALIZED;
	hydrogenInstance->infoLog( "UNINITIALIZED" );

	// raise stateChanged event
	for (uint i = 0; i < audioEngine_listeners.size(); i++) {
		audioEngine_listeners[i]->stateChanged(UNINITIALIZED);
	}


	// delete all copied notes in the song notes queue
	for (uint i = 0; i < songNoteQueue.size(); i++) {
		Note *note = songNoteQueue[i];
		delete note;
		note = NULL;
	}
	songNoteQueue.clear();

	// delete all copied notes in the playing notes queue
	for (uint i = 0; i < playingNotesQueue.size(); i++) {
		Note *note = playingNotesQueue[i];
		delete note;
		note = NULL;
	}
	playingNotesQueue.clear();

	// delete all copied notes in the midi notes queue
	for (uint i = 0; i < midiNoteQueue.size(); i++) {
		Note *note = midiNoteQueue[i];
		delete note;
		note = NULL;
	}
	midiNoteQueue.clear();

	hydrogenInstance->infoLog( "audioEngine_destroy: all notes buffers cleared" );

	hydrogenInstance->infoLog("Destroying delayBuffer_L");
	delete[] delayBuffer_L;
	delayBuffer_L = NULL;

	hydrogenInstance->infoLog("Destroying delayBuffer_R");
	delete[] delayBuffer_R;
	delayBuffer_R = NULL;

	delete metronomeInstrument;
	metronomeInstrument = NULL;

	audioEngine_unlock();
}





//----------------------------------------------------------------------------
/**
 * Start playing
 * return 0 = OK
 * return -1 = NULL Audio Driver
 * return -2 = Driver connect() error
 */
int audioEngine_start(bool bLockEngine) {
	if (bLockEngine) {
		audioEngine_lock();
	}

	hydrogenInstance->infoLog("Hydrogen audio engine start");

	// check current state
	if (audioEngine_state != READY) {
		hydrogenInstance->errorLog("audioEngine_start: Error the audio engine is not in READY state");
		if (bLockEngine) {
			audioEngine_unlock();
		}
		return 0;	// FIXME!!
	}


	PreferencesMng *preferencesMng = PreferencesMng::getInstance();
	masterPeak_L = 0.0f;
	masterPeak_R = 0.0f;
	delayBufferPeak_L = 0.0f;
	delayBufferPeak_R = 0.0f;
	totalFrames = 0;	// reset total frames
	patternPos = -1;
	m_nPatternTickPosition = 0;
	patternStartTick = -1;

	// reset
	findPatternInTick_lastTotal = 0;
	findPatternInTick_lastTick = 0;

	// prepare the tickSize for this song
/*
	float bpm = song->getBpm();
	float resolution = song->getResolution();
	float sampleRate = (float)driver->getSampleRate();
	tickSize = (uint)( sampleRate * 60.0 /  bpm / resolution );
*/
	updateTickSize();


	maxNotes = preferencesMng->getMaxNotes();

	// set initial metronome volume
	if ( preferencesMng->isMetronomeEnabled() ) {
		metronomeInstrument->setVolume( preferencesMng->getMetronomeVolume() );
	}
	else {
		metronomeInstrument->setVolume( 0.0 );
	}

	// change the current audio engine state
	audioEngine_state = PLAYING;
	hydrogenInstance->infoLog("PLAYING");
	// raise stateChanged event
	for (uint i = 0; i < audioEngine_listeners.size(); i++) {
		audioEngine_listeners[i]->stateChanged(PLAYING);
	}

	if (bLockEngine) {
		audioEngine_unlock();
	}
	return 0; // per ora restituisco sempre OK
}





//----------------------------------------------------------------------------
/**
 * Stop the audio engine
 */
void audioEngine_stop(bool bLockEngine) {
	if (bLockEngine) {
		audioEngine_lock();
	}
	hydrogenInstance->infoLog("Hydrogen audio engine stop");

	// check current state
	if (audioEngine_state != PLAYING) {
		hydrogenInstance->errorLog("audioEngine_stop: Error the audio engine is not in PLAYING state");
		cout << "state = " << audioEngine_state << endl;
		if (bLockEngine) {
			audioEngine_unlock();
		}
		return;
	}


	// change the current audio engine state
	audioEngine_state = READY;
	hydrogenInstance->infoLog("READY");
	// raise stateChanged event
	for (uint i = 0; i < audioEngine_listeners.size(); i++) {
		audioEngine_listeners[i]->stateChanged(READY);
	}

	masterPeak_L = 0.0f;
	masterPeak_R = 0.0f;
	delayBufferPeak_L = 0.0f;
	delayBufferPeak_R = 0.0f;
	m_nPatternTickPosition = 0;
	patternStartTick = -1;

	// delete all copied notes in the song notes queue
	for (uint i = 0; i < songNoteQueue.size(); i++) {
		Note *note = songNoteQueue[i];
		if (note) {
			delete note;
			note = NULL;
		}
	}
	songNoteQueue.clear();

/*
	// delete all copied notes in the playing notes queue
	for (uint i = 0; i < playingNotesQueue.size(); i++) {
		Note *note = playingNotesQueue[i];
		if (note) {
			delete note;
			note = NULL;
		}
	}
	playingNotesQueue.clear();
*/

	// delete all copied notes in the midi notes queue
	for (uint i = 0; i < midiNoteQueue.size(); i++) {
		Note *note = midiNoteQueue[i];
		if (note) {
			delete note;
			note = NULL;
		}
	}
	midiNoteQueue.clear();

	if (bLockEngine) {
		audioEngine_unlock();
	}
}




//----------------------------------------------------------------------------
/**
 *
 */
inline void audioEngine_process_checkBPMChanged() {


	float bpm = song->getBpm();
	float resolution = song->getResolution();
	float sampleRate = (float)driver->getSampleRate();
	uint newTickSize = (uint)(sampleRate * 60.0 /  bpm / resolution);

	if (newTickSize != tickSize) {
		// update tickSize
		tickSize = newTickSize;
		hydrogenInstance->infoLog("[audioEngine_process_checkBPMChanged()] tickSize changed");
		audioEngine_clearNoteQueue();
	}

}





//----------------------------------------------------------------------------
/**
 *
 */
inline void audioEngine_process_playPattern( unsigned long nframes )
{
	if ( audioEngine_state == PLAYING ) {
		// leggo da songNoteQueue
		while (songNoteQueue.size() > 0) {
			Note *note = songNoteQueue[0];

			// verifico se la nota rientra in questo ciclo
			uint noteStartInFrames = note->getPosition() * tickSize;

			// totalFrames <= NotePos < totalFrames + bufferSize
			bool isNoteStart = ( ( noteStartInFrames >= totalFrames ) && ( noteStartInFrames < ( totalFrames + nframes ) ) );
			bool isOldNote = noteStartInFrames < totalFrames;
			if ( isNoteStart || isOldNote ) {
				playingNotesQueue.push_back( note );		// aggiungo la nota alla lista di note da eseguire
				songNoteQueue.pop_front();			// rimuovo la nota dalla lista di note

				/* TEST exclusive notes */

				Instrument *instr = note->getInstrument();
				vector<Instrument*> excludeVect = instr->getExcludeVect();
				if ( excludeVect.size() != 0 ) {
					// for
					Instrument *instr = excludeVect[ 0 ];
					// remove all notes listed in excludeVect
					for ( uint i = 0; i < playingNotesQueue.size(); i++ ) {	// delete older note
						Note *oldNote = playingNotesQueue[ i ];

						if ( oldNote->getInstrument() == instr ) {
							cout << "Delete excluded note" << endl;
							playingNotesQueue.erase( playingNotesQueue.begin() + i );
							delete oldNote;
						}
					}

				}

				/* end TEST exclusive notes */



				// raise noteOn event
				for ( uint nList = 0; nList < audioEngine_listeners.size(); nList++ ) {
					audioEngine_listeners[nList]->noteOn( note->copy() );	// the note must be deleted!
				}
				continue;
			}
			else {
				// la nota non andra' in esecuzione
				break;
			}
		}
	}

}




//----------------------------------------------------------------------------
/**
 *
 */
inline void audioEngine_process_playMidi( unsigned long nframes )
{
	// leggo da midiNoteQueue
	while ( midiNoteQueue.size() > 0 ) {
		Note *note = midiNoteQueue[0];
		playingNotesQueue.push_back( note );		// aggiungo la nota alla lista di note da eseguire
		midiNoteQueue.pop_front();			// rimuovo la nota dalla lista di note

		/* TEST exclusive notes */
		Instrument *instr = note->getInstrument();
		vector<Instrument*> excludeVect = instr->getExcludeVect();
		if ( excludeVect.size() != 0 ) {
			Instrument *instr = excludeVect[ 0 ];
			// remove all notes listed in excludeVect
			for ( uint i = 0; i < playingNotesQueue.size(); i++ ) {	// delete older note
				Note *oldNote = playingNotesQueue[ i ];

				if ( oldNote->getInstrument() == instr ) {
					cout << "cancello" << endl;
					playingNotesQueue.erase( playingNotesQueue.begin() + i );
					delete oldNote;
				}
			}
		}
		/* end TEST exclusive notes */

		// raise noteOn event
		for ( uint nList = 0; nList < audioEngine_listeners.size(); nList++ ) {
			audioEngine_listeners[nList]->noteOn( note->copy() );	// the note must be deleted!
		}
	}
}





//----------------------------------------------------------------------------
/**
 * Update engine info
 */
inline void audioEngine_process_updateInfo() {
	playingNotes = playingNotesQueue.size();
}




/**
 *
 */
void audioEngine_seek( unsigned long nFrames )
{
	char tmpBuf[100];
	sprintf(tmpBuf, "[audioEngine_seek()] seek in %d", nFrames);
	hydrogenInstance->infoLog(tmpBuf);

	totalFrames = nFrames;

	int tickNumber_start = (uint)( totalFrames / tickSize );

	patternStartTick = -1;
	patternPos = findPatternInTick( tickNumber_start, true );

	if ( song->getMode() == LIVE_MODE ) {
		nextPattern = song->getPatternSequence()->get( patternPos );
	}

	audioEngine_clearNoteQueue();
}



/**
 *
 */
/**
 *
 */
inline void audioEngine_process_jackTransport()
{
#ifdef JACK_SUPPORT

	// skip if jack transport is disabled
	if ( (PreferencesMng::getInstance())->isJackTransportSlave() == false ) {
		return;
	}

	static double last_jack_bpm = 0;

	if ( driver->getClassName() == "JackDriver" ) {
		JackDriver *jackDriver = ( (JackDriver*)driver );
		jackDriver->updateTransportInfo();

		jack_transport_state_t state= jackDriver->getTransportState();
		jack_position_t position = jackDriver->getTransportPos();

		switch (state) {
			case JackTransportStopped:
				// check current state
				if (audioEngine_state == PLAYING) {
					hydrogenInstance->infoLog( "JackTransportStopped" );

					audioEngine_stop(false);	// no engine lock
					audioEngine_seek( position.frame );
					audioEngine_clearNoteQueue();
				}
				break;

			case JackTransportRolling:
				if (audioEngine_state == READY) {	// check current state
					if ( totalFrames != position.frame ) {
						hydrogenInstance->infoLog("[audioEngine_process_jackTransport()] Rolling");

						audioEngine_start(false);	// no engine lock

						if (position.valid & JackPositionBBT) {
							last_jack_bpm = position.beats_per_minute;

							#warning "Change song bpm when jack bpm changes"
							double bpm = position.beats_per_minute;
							double resolution = song->getResolution();
							double sampleRate = (float)driver->getSampleRate();
							tickSize = (uint)( sampleRate * 60.0 /  bpm / resolution );

							char tmp[20];
							sprintf(tmp, "[audioEngine_process_jackTransport()] Rolling, ticksize = %d", tickSize);
							hydrogenInstance->infoLog(tmp);
						}
						else{
							hydrogenInstance->warningLog( "JackPosition not valid" );
						}
						audioEngine_seek( position.frame );
						char tmp[20];
						sprintf(tmp, "[audioEngine_process_jackTransport()] Rolling, totalFrames = %d", totalFrames);
						hydrogenInstance->infoLog(tmp);
					}
				}
				else if ( audioEngine_state == PLAYING ) {
					if ( totalFrames != position.frame ) {
						hydrogenInstance->infoLog("[audioEngine_process_jackTransport()] Rolling");
						cout << "audioEngine_process_jackTransport. totalFrames = " << totalFrames << endl;
						audioEngine_seek( position.frame );
						audioEngine_clearNoteQueue();
					}
					if (position.valid & JackPositionBBT) {
						if ( ((double)last_jack_bpm != (double)position.beats_per_minute) && (position.beats_per_minute != 0) ) {
							hydrogenInstance->infoLog("[audioEngine_process_jackTransport()] Rolling");

							audioEngine_stop(false);	// no engine lock
							audioEngine_start(false);	// no engine lock
							cout << "old ticksize = " << tickSize << endl;
							cout << "new bpm = " << position.beats_per_minute << endl;

							double bpm = position.beats_per_minute;
							float resolution = song->getResolution();
							float sampleRate = (float)driver->getSampleRate();
							tickSize = (uint)( sampleRate * 60.0 /  bpm / resolution );

							last_jack_bpm = bpm;
							cout << "new jack ticksize = " << tickSize << endl;
						}
					}
				}
				break;

			case JackTransportStarting:
				hydrogenInstance->infoLog("[audioEngine_process_jackTransport()] JackTransportStarting not implemented yet");
				break;

			default:
				hydrogenInstance->errorLog( "[audioEngine_process_jackTransport()] Unknown jack transport state" );
		}
	}

#endif // #ifdef JACK_SUPPORT

}


void audioEngine_clearNoteQueue() {
	hydrogenInstance->infoLog("[audioEngine_clearNoteQueue()]");

	// delete all copied notes in the song notes queue
	for (uint i = 0; i < songNoteQueue.size(); i++) {
		Note *note = songNoteQueue[i];
		delete note;
//		note = NULL;
//		songNoteQueue[i] = NULL;
	}
	songNoteQueue.clear();


	// delete all copied notes in the playing notes queue
	for (uint i = 0; i < playingNotesQueue.size(); i++) {
		Note *note = playingNotesQueue[i];
		delete note;
//		note = NULL;
//		playingNotesQueue[i] = NULL;
	}
	playingNotesQueue.clear();


	// delete all copied notes in the midi notes queue
	for (uint i = 0; i < midiNoteQueue.size(); i++) {
		Note *note = midiNoteQueue[i];
		delete note;
//		note = NULL;
//		midiNoteQueue[i] = NULL;
	}
	midiNoteQueue.clear();

}





//----------------------------------------------------------------------------
/**
 *
 */
int audioEngine_process(uint32_t nframes, void *arg)
{
//	audioEngine_lock();

	// try to lock the engine mutex
	// if fails return without sound processing
	int res = pthread_mutex_trylock(&engineLock_mutex);
	if (res != 0) {
		// update total frames number
		totalFrames += nframes;
		hydrogenInstance->warningLog( "audioEngine_process: trylock != 0" );
		return 0;
	}


	timeval startTimeval = currentTime2();


	audioEngine_process_jackTransport();

	// (re)create audio buffer
	if ( audioEngine_buffer_size != nframes ) {
		hydrogenInstance->warningLog( "buffer size changed" );
		cout << "audioEngine_buffer_size = " << audioEngine_buffer_size << endl;
		cout << "nframes = " << nframes << endl;

		if ( delayBuffer_L != NULL ) {
			hydrogenInstance->infoLog( "audioEngine_process: Destroying delayBuffer_L" );
			delete[] delayBuffer_L;
			delayBuffer_L = NULL;
		}
		if ( delayBuffer_R != NULL ) {
			hydrogenInstance->infoLog( "audioEngine_process: Destroying delayBuffer_R" );
			delete[] delayBuffer_R;
			delayBuffer_R = NULL;
		}
		audioEngine_buffer_size = nframes;

		hydrogenInstance->infoLog( "audioEngine_process: Creating delayBuffer_L" );
		delayBuffer_L = new float[ nframes ];
		hydrogenInstance->infoLog( "audioEngine_process: Creating delayBuffer_R" );
		delayBuffer_R = new float[ nframes ];
	}

	memset( main_buffer_L, 0, nframes * sizeof( float ) );		// clear main out Left
	memset( main_buffer_R, 0, nframes * sizeof( float ) );	// clear main out Right

	if ( ( audioEngine_state == READY ) || ( audioEngine_state == PLAYING) ) {
		if ( song->isDelayEnabled() ) {
			memset( delayBuffer_L, 0, nframes * sizeof( float ) );		// clear delay buffer Left
			memset( delayBuffer_R, 0, nframes * sizeof( float ) );		// clear delay buffer Right
		}
	}

	if (use_track_outs) {
		// clear track buffers
	 	for (unsigned int n=0; n < MAX_TRACKS; ++n) {
 			memset( track_buffers_L[n], 0, nframes * sizeof( float ) );
 			memset( track_buffers_R[n], 0, nframes * sizeof( float ) );
	 	}
	}

	if ( ( audioEngine_state == READY) || (audioEngine_state == PLAYING ) ) {
		audioEngine_process_checkBPMChanged();
	}

	// check ticksize
	if ( tickSize == 0 ) {
		hydrogenInstance->infoLog( "audioEngine_process: tickSize == 0" );
		if ( audioEngine_state != PREPARED ) {
			hydrogenInstance->errorLog( "tickSize == 0 audioEngine_state != PREPARED!" );
		}
	}
	audioEngine_process_updateInfo();

	bool sendPatternChange = false;
	if ( audioEngine_state == PLAYING ) {
		// update the notes queue
		int res = audioEngine_updateNoteQueue( nframes );
		if ( res == -1 ) {	// end of song
			hydrogenInstance->infoLog( "audioEngine_process: End of song received, calling engine_stop()" );
			audioEngine_unlock();
			audioEngine_stop();
			if ( driver->getClassName() == "DiskWriterDriver" ) {
				hydrogenInstance->infoLog( "Exiting DiskWriterDriver thread!!" );
				return 1;	// kill the audio driver thread
			}
		}
		else if ( res == 2 ) {	// send pattern change
			sendPatternChange = true;
		}
	}
	// play PATTERN
	audioEngine_process_playPattern( nframes );

	// MIDI EVENTS
	if ( ( audioEngine_state == READY ) || ( audioEngine_state == PLAYING) ) {
		audioEngine_process_playMidi(nframes);
	}

	// Max notes
	while ( playingNotesQueue.size() > maxNotes ) {
		Note *oldNote = playingNotesQueue[ 0 ];
		playingNotesQueue.erase( playingNotesQueue.begin() );
		delete oldNote;
	}


	// eseguo tutte le note nella lista di note in esecuzione
	uint i = 0;
	while ( i < playingNotesQueue.size() ) {
		Note* note = playingNotesQueue[ i ];		// recupero una nuova nota
		uint res = audioEngine_renderNote( note, nframes, nframes );
		if ( res == 1 ) {	// la nota e' finita
			playingNotesQueue.erase( playingNotesQueue.begin() + i );
			if ( note != NULL ) {
				delete note;
				note = NULL;
			}
		}
		else {
			i++; // carico la prox nota
		}
	}


	if ( ( audioEngine_state == READY ) || ( audioEngine_state == PLAYING) ) {
		float delay_wetLevel = song->getDelayWetLevel();
		uint delay_delayTime = tickSize * song->getDelayTime() * 3;
		float delay_feedbackLevel = song->getDelayFeedbackLevel();

		for ( uint i = 0; i < nframes; i++) {
			float val_L = main_buffer_L[i];
			float val_R = main_buffer_R[i];

			// Delaly FX
			if ( song->isDelayEnabled() ) {

				if (delay_delayTime > MAX_DELAY) {
					hydrogenInstance->errorLog( "audioEngine_process: delayTime > MAX_DELAY" );
					delay_delayTime = MAX_DELAY;
				}
				int currentDelayPos = (delayStartPos + i) % delay_delayTime;

				float delayVal_L = delayRingBuffer_L[ currentDelayPos ]  * delay_wetLevel;
				float delayVal_R = delayRingBuffer_R[ currentDelayPos ] * delay_wetLevel;

				val_L += delayVal_L;
				val_R += delayVal_R;

				// feedback
				float temp_L = delayBuffer_L[i] + ( delayRingBuffer_L[ currentDelayPos ] * delay_feedbackLevel );
				float temp_R = delayBuffer_R[i] + ( delayRingBuffer_R[ currentDelayPos ] * delay_feedbackLevel );

				bool delay_pingPong = true;
				// ping pong delay
				if ( delay_pingPong == true ) {
					delayRingBuffer_R[ currentDelayPos ] = temp_L;
					delayRingBuffer_L[ currentDelayPos ] = temp_R;
				}
				else {
					delayRingBuffer_L[ currentDelayPos ] = temp_L;
					delayRingBuffer_R[ currentDelayPos ] = temp_R;
				}

				// update delayBuffer peak
				if ( delayVal_L > delayBufferPeak_L ) {
					delayBufferPeak_L = delayVal_L;
				}
				if ( delayVal_R > delayBufferPeak_R ) {
					delayBufferPeak_R = delayVal_R;
				}
			}

			main_buffer_L[i] = val_L;
			main_buffer_R[i] = val_R;

			// update master peak
			if (val_L > masterPeak_L) {
				masterPeak_L = val_L;
			}
			if (val_R > masterPeak_R) {
				masterPeak_R = val_R;
			}
		}
		delayStartPos = (delayStartPos + nframes) % delay_delayTime;
	}


	// update total frames number
	totalFrames += nframes;

	timeval finishTimeval = currentTime2();
	processTime = (finishTimeval.tv_sec - startTimeval.tv_sec) * 1000.0 + (finishTimeval.tv_usec - startTimeval.tv_usec) / 1000.0;

	float sampleRate = (float)driver->getSampleRate();
	maxProcessTime = 1000.0 / (sampleRate / nframes);

#ifdef DEBUG_MESSAGES
	//DEBUG
	if ( processTime > maxProcessTime ) {
		char xrunMsg[200];
		sprintf( xrunMsg, "XRUN of %f msec ( %f > %f )", (processTime - maxProcessTime), processTime, maxProcessTime);
		hydrogenInstance->warningLog( string( xrunMsg ) );

		char xrunInfo[200];
		sprintf( xrunInfo, "Playing notes = %d", playingNotesQueue.size() );
		hydrogenInstance->warningLog( string( xrunInfo ) );
	}
#endif

	audioEngine_unlock();

	if ( sendPatternChange ) {
		// raise patternChanged event
		for (uint i = 0; i < audioEngine_listeners.size(); i++) {
			audioEngine_listeners[i]->patternChanged();
		}
	}

	return 0;
}





//----------------------------------------------------------------------------
/** Funzione per il rendering di una singola nota
 *  Restituisce 0 se la nota non e' finita
 *  Restituisce 1 se la nota e' terminata
 */
inline uint audioEngine_renderNote(Note* note, int nFrames, uint bufferSize) {
	// nFrames = frame disponibili da copiare nei buffer

/*
	if (note->samplePosition == 0) {
		hydrogenInstance->infoLog( "note on" );
	}
*/
	Instrument *instr = note->getInstrument();
	if (instr == NULL) {
		hydrogenInstance->errorLog( "audioEngine_renderNote: Instrument Null" );
	}

	int track_n = song->getInstrumentList()->getPos(instr);

	Sample *s = instr->getSample();
	if (s == NULL) {
		hydrogenInstance->errorLog("audioEngine_renderNote: sample Null");
	}

	float *data_L = s->getData_L();
	float *data_R = s->getData_R();
	float instrPeak_L = instr->getPeak_L(); // this value will be reset to 0 by the mixer..
	float instrPeak_R = instr->getPeak_R(); // this value will be reset to 0 by the mixer..

	int noteStartInFrames = ( note->getPosition() * tickSize ) + note->getHumanizeDelay();

	int initialSilence = 0;
	if (noteStartInFrames > (int)totalFrames) {	// scrivo silenzio prima dell'inizio della nota
		initialSilence = noteStartInFrames - totalFrames;
		nFrames = nFrames - initialSilence;
		if (nFrames < 0) {	// delay note execution
			return 0;
		}

	}

	// verifico il numero di frame disponibili ancora da eseguire
	int avail_bytes = s->getNFrames() - note->samplePosition;

	float cost_L = 1;
	float cost_R = 1;
	float cost_track = 1;

	if (instr->isMuted()) {	// is instrument muted?
		cost_L = 0.0;
		cost_R = 0.0;
	}
	else {	// Precalcolo i valori per le nota
		cost_L = cost_L * note->getVelocity();		// note velocity
		cost_L = cost_L * note->getPan_L();		// note pan
		cost_L = cost_L * instr->getPan_L();		// instrument pan
		cost_L = cost_L * instr->getVolume();		// instrument volume
		cost_L = cost_L * song->getVolume();		// song volume

		cost_R = cost_R * note->getVelocity();	// note velocity
		cost_R = cost_R * note->getPan_R();		// note pan
		cost_R = cost_R * instr->getPan_R();		// instrument pan
		cost_R = cost_R * instr->getVolume();		// instrument volume
		cost_R = cost_R * song->getVolume();	// song pan
	}

	// direct track outputs only use velocity
	cost_track = note->getVelocity();

	float delayLevel = instr->getDelayLevel();

	// il sample e' piu' grande del buffersize
	if (avail_bytes > (int)(bufferSize - initialSilence)) {
		// imposto il numero dei bytes disponibili uguale al buffersize
		avail_bytes = bufferSize - initialSilence;

		// copio il sample nel buffer
		for (int i = 0; i < avail_bytes; i++) {
			int samplePos = note->samplePosition + i;

			float val_L = data_L[samplePos] * cost_L;
			float val_R = data_R[samplePos] * cost_R;

			int bufferPos = initialSilence + i;

			main_buffer_L[bufferPos] += val_L;
			main_buffer_R[bufferPos] += val_R;

			delayBuffer_L[bufferPos] += val_L * delayLevel;
			delayBuffer_R[bufferPos] += val_R * delayLevel;

			// for the track output
			if (use_track_outs && track_n >= 0) {
				track_buffers_L[track_n][bufferPos] = data_L[samplePos] * cost_track;
				track_buffers_R[track_n][bufferPos] = data_R[samplePos] * cost_track;
			}

			// update instr peak
			if (val_L > instrPeak_L) {	instrPeak_L = val_L;	}
			if (val_R > instrPeak_R) {	instrPeak_R = val_R;	}
		}
		instr->setPeak_L( instrPeak_L );
		instr->setPeak_R( instrPeak_R );

		// aggiorno la posizione del sample nella nota
		note->samplePosition += avail_bytes;
		return 0;	// la nota non e' ancora terminata
	}
	else {	// copio i frames disponibili nel buffer

		// copio il sample nel buffer
		for (int i = 0; i < avail_bytes; i++) {
			int samplePos = note->samplePosition + i;
			float val_L = data_L[samplePos] * cost_L;
			float val_R = data_R[samplePos] * cost_R;

			int bufferPos = (initialSilence) + i;

			main_buffer_L[bufferPos] += val_L;
			main_buffer_R[bufferPos] += val_R;

			delayBuffer_L[bufferPos] += val_L * delayLevel;
			delayBuffer_R[bufferPos] += val_R * delayLevel;

			// for the track output
			if (use_track_outs && track_n >= 0) {
				track_buffers_L[track_n][bufferPos] = data_L[samplePos] * cost_track;
				track_buffers_R[track_n][bufferPos] = data_R[samplePos] * cost_track;
			}

			// update instr peak
			if (val_L > instrPeak_L) {	instrPeak_L = val_L;	}
			if (val_R > instrPeak_R) {	instrPeak_R = val_R;	}
		}
		instr->setPeak_L( instrPeak_L );
		instr->setPeak_R( instrPeak_R );

		return 1;	// la nota e' finita
	}
	return 0;	// not reached
}





//----------------------------------------------------------------------------
/**
 *
 */
void audioEngine_setSong(Song *newSong) {
	hydrogenInstance->infoLog( "set song \"" + newSong->getName() + "\"" );

	audioEngine_lock();

	// check current state
	if (audioEngine_state != PREPARED) {
		hydrogenInstance->errorLog("audioEngine_setSong: Error the audio engine is not in PREPARED state");
	}

	pattern = NULL;
	nextPattern = NULL;

	audioEngine_clearNoteQueue();

	hydrogenInstance->infoLog( "audioEngine_setSong: all notes buffers cleared" );

	song = newSong;

	// update ticksize
	audioEngine_process_checkBPMChanged();

	// find the first pattern and set as current
	if ( song->getPatternList()->getSize() > 0 ) {
		pattern = song->getPatternList()->get(0);
	}

	// change the current audio engine state
	audioEngine_state = READY;
	hydrogenInstance->infoLog("READY");

	// raise patternChanged event
	for (uint i = 0; i < audioEngine_listeners.size(); i++) {
		audioEngine_listeners[i]->patternChanged();
	}

	// raise stateChanged event
	for (uint i = 0; i < audioEngine_listeners.size(); i++) {
		audioEngine_listeners[i]->stateChanged(READY);
	}

	audioEngine_unlock();
}



//----------------------------------------------------------------------------
/**
 *
 */
void audioEngine_removeSong() {
	audioEngine_lock();
	// check current state
	if (audioEngine_state != READY) {
		hydrogenInstance->errorLog("audioEngine_setSong: Error the audio engine is not in READY state");
		audioEngine_unlock();
		return;
	}

	song = NULL;
	pattern = NULL;
	nextPattern = NULL;


	audioEngine_clearNoteQueue();

	// change the current audio engine state
	audioEngine_state = PREPARED;
	// raise stateChanged event
	for (uint i = 0; i < audioEngine_listeners.size(); i++) {
		audioEngine_listeners[i]->stateChanged(PREPARED);
	}

	audioEngine_unlock();
}




//----------------------------------------------------------------------------
/**
 *
 */
// return -1 = end of song
// return 2 = send pattern changed event!!
inline int audioEngine_updateNoteQueue(uint nFrames)
{
	static int lastTick = -1;
	bool sendPatternChange = false;

	int tickNumber_start = (uint)( totalFrames / tickSize );
	int tickNumber_end = (uint)( ( totalFrames + nFrames ) / tickSize );

	int tick = tickNumber_start;
	while ( tick <= tickNumber_end ) {
		if (tick == lastTick) {
			tick++;
			continue;
		}
		else {
			lastTick = tick;
		}
//		cout << "\ntick " << tick << endl;
//		cout << "pat tick = " << m_nPatternTickPosition << endl;

		// SONG MODE
		if ( song->getMode() == SONG_MODE ) {

			uint patternSize = pattern->getSize();
			if ( ( tick >= (int)( patternStartTick + patternSize ) ) || (patternStartTick == -1) ) {
				bool loop = song->isLoopEnabled();
				patternPos = findPatternInTick( tick, loop );
				patternStartTick = tick;
				sendPatternChange = true;
			}

			m_nPatternTickPosition = tick - patternStartTick;

			if ( patternPos == -1 ) {
				cout << "*** pattern pos = -1 ***" << endl;
				// check if the song loop is enabled
				if ( song->isLoopEnabled() == true ) {
					if ( song->getPatternSequence()->getSize() != 0 ) {
						if ( patternPos == -1 ) {
							cout << "LOOP" << endl;
						}
					}
				}
				else {
					hydrogenInstance->infoLog( "End of Song" );
					return -1;
				}
			}
			pattern = song->getPatternSequence()->get( patternPos );
		}

		// LIVE MODE
		else if ( song->getMode() == LIVE_MODE )	{
			if ( ( tick == (int)(patternStartTick + pattern->getSize()) ) || ( patternStartTick == -1 ) ) {
				if ( nextPattern != NULL ) {
					pattern = nextPattern;
					nextPattern = NULL;
					sendPatternChange = true;
				}
				patternStartTick = tick;
			}
			m_nPatternTickPosition = tick - patternStartTick;
		}

		// metronome
		if ( ( tick % 48 ) == 0) {
			if (metronomeInstrument->getVolume() != 0) {
				Note *metronomeNote = new Note( tick, 1.0, 1.0, 1.0 );
				metronomeNote->setInstrument( metronomeInstrument );
				audioEngine_noteOn( metronomeNote );
			}
		}

		// update notes queue
		if ( pattern != NULL ) {
			uint nSequences = ( pattern->getSequenceList() )->getSize();
			//uint nNote = tick % MAX_NOTES;
			uint nNote = m_nPatternTickPosition;

			for ( uint i = 0; i < nSequences; i++ ) {
				Sequence* sequence = ( (pattern->getSequenceList() )->get(i) );
				Note* note = sequence->noteList[nNote];
				if ( note ) {
					uint offset = 0;
					float velocity = note->getVelocity();

					// Swing
					if ( song->isSwingEnabled() ) {
						float swingFactor = song->getSwingFactor();
						if ( ( (nNote % 12) == 0 ) && ( (nNote % 24) != 0) ) {	// da l'accento al tick 4, 12, 20, 36...
							offset += (int)( ( 2.0 * tickSize ) * swingFactor );
						}
					}

					// Humanize - Time parameter
					if ( song->isHumanizeTimeEnabled() ) {
						int maxTimeHumanize = 2000;
						if (song->getHumanizeTimeValue() != 0) {
							float random = randomValue(100) / 100.0;
							offset += (int)(random * song->getHumanizeTimeValue() * maxTimeHumanize);
						}
					}

					// Humanize - Velocity parameter
					if ( song->isHumanizeVelocityEnabled() ) {
						if (song->getHumanizeVelocityValue() != 0) {
							float random = song->getHumanizeVelocityValue() * (randomValue(100) / 100.0);
							velocity += (random - (song->getHumanizeVelocityValue() / 2.0));
							if (velocity > 1.0) {
								velocity = 1.0;
							}
							else if (velocity < 0.0) {
								velocity = 0.0;
							}
						}
					}

					Note *copiedNote = new Note( note->getPosition(), velocity, note->getPan_L(), note->getPan_R() );
					copiedNote->setInstrument( note->getInstrument() );
					copiedNote->setPosition( tick );
					copiedNote->samplePosition = 0;
					copiedNote->setHumanizeDelay( offset );	// humanize time
					songNoteQueue.push_back( copiedNote );
				}
			}
		}

		tick++;
	}


	// audioEngine_process must send the pattern change event after mutex unlock
	if (sendPatternChange)
	{
		return 2;
	}
	return 0;
}




/**
 * Find the pattern for tick number
 */
inline int findPatternInTick( int tick, bool loopMode )
{
//	cout << "findPatternInTick " << tick << endl;
	if (song == NULL) {
		return -1;
	}

//	cout << "find pattern in tick " << tick << " [lastTick=" << findPatternInTick_lastTick << "] lastTotal[" << findPatternInTick_lastTotal << "] "	 << endl;
	int pos = -1;

	int totalTick = 0;
	int nPatterns = song->getPatternSequence()->getSize();
	int patternSize = 0;


/*
	// vecchia ricerca
	for ( int i = 0; i < nPatterns; i++ ) {
		cout << i << endl;
		patternSize = song->getPatternSequence()->get( i )->getSize();
		if ( (tick >= totalTick) && ( tick < totalTick + patternSize ) ) {
			cout << endl;
			findPatternInTick_lastTick = i;
			return i;
		}
		totalTick += patternSize;
	}


	cout << "nuovo.." << endl;
*/

	totalTick = findPatternInTick_lastTotal;
	// nuova ricerca semi-ottimizzata
	for ( int i = findPatternInTick_lastTick; i < nPatterns; i++ ) {
//		cout << i << endl;
		Pattern *pattern = song->getPatternSequence()->get( i );
		if (pattern == NULL) {
			cout << "Hydrogen::findPatternInTick. Pattern == null" << endl;
			return -1;
		}
		patternSize = pattern->getSize();

		if ( (tick >= totalTick) && ( tick < totalTick + patternSize ) ) {
			findPatternInTick_lastTick = i;
			findPatternInTick_lastTotal = totalTick;
			return i;
		}
		totalTick += patternSize;
	}
/*
	for ( int i = 0; i < lastTick; i++ ) {
		cout << i << endl;
		Pattern *pattern = song->getPatternSequence()->get( i );
		if (pattern == NULL) {
			return -1;
		}
		patternSize = pattern->getSize();

		if ( (tick >= totalTick) && ( tick < totalTick + patternSize ) ) {
			lastTotal = totalTick;
			lastTick = i;
			return i;
		}

		totalTick += patternSize;
	}
*/

	if ( loopMode ) {
		cout << "loop mode" << endl;
		// se arriva qua e' attivo il loop mode
		int songTick = totalTick;

		int loopTick = 0;
		if (songTick != 0) {
			loopTick = tick % songTick;
		}

		totalTick = 0;
		for ( int i = 0; i < nPatterns; i++ ) {
			patternSize = song->getPatternSequence()->get( i )->getSize();
			if ( ( loopTick >= totalTick ) && ( loopTick < totalTick + patternSize ) ) {
				findPatternInTick_lastTick = i;
				findPatternInTick_lastTotal = totalTick;
				return i;
			}

			totalTick += patternSize;
		}
	}

	findPatternInTick_lastTick = 0;
	totalTick = 0;
	return pos;
}




//----------------------------------------------------------------------------
/**
 *
 */
void	audioEngine_noteOn(Note *note) {
	// check current state
	if ( (audioEngine_state != READY) && (audioEngine_state != PLAYING) ) {
		hydrogenInstance->errorLog("audioEngine_noteOn: Error the audio engine is not in READY state");
		delete note;
		return;
	}

	midiNoteQueue.push_back(note);
}




//----------------------------------------------------------------------------
/**
 *
 */
void	audioEngine_noteOff( Note *note )
{
	if (note == NULL)
	{
		hydrogenInstance->errorLog("audioEngine_noteOff: Error, note == NULL");
	}

	audioEngine_lock();

	// check current state
	if ( (audioEngine_state != READY) && (audioEngine_state != PLAYING) )
	{
		hydrogenInstance->errorLog("audioEngine_noteOff: Error the audio engine is not in READY state");
		delete note;
		audioEngine_unlock();
		return;
	}

	for ( uint i = 0; i < playingNotesQueue.size(); i++ ) {	// delete old note
		Note *oldNote = playingNotesQueue[ i ];

		if ( oldNote->getInstrument() == note->getInstrument() ) {
			playingNotesQueue.erase( playingNotesQueue.begin() + i );
			delete oldNote;
			break;
		}
	}
	audioEngine_unlock();

	delete note;
}




//----------------------------------------------------------------------------
/**
 *
 */
unsigned long audioEngine_getTickPosition() {
	return m_nPatternTickPosition;
}





//----------------------------------------------------------------------------
/**
 * Start all audio drivers
 */
void audioEngine_startAudioDrivers() {
	audioEngine_lock();

	hydrogenInstance->infoLog("Starting audio drivers");

	// check current state
	if (audioEngine_state != INITIALIZED) {
		hydrogenInstance->errorLog("audioEngine_startAudioDrivers: Error the audio engine is not in INITIALIZED state");
		return;
	}

	// check if the audio driver is still alive
	if (driver) {
		hydrogenInstance->errorLog( "The audio driver is still alive" );
	}


	PreferencesMng *preferencesMng = PreferencesMng::getInstance();

	use_track_outs = false;

	// Create Audio Driver
	string audioDriver = preferencesMng->getAudioDriver();
	if (audioDriver == "Oss") {
#ifdef OSS_SUPPORT
		driver = new OssDriver( audioEngine_process );
#else
		audioEngine_raiseError( NO_OSS_SUPPORT );
		driver = new NullDriver( audioEngine_process );
#endif
	}
	else if (audioDriver == "Jack") {
#ifdef JACK_SUPPORT
		driver = new JackDriver(audioEngine_process);

		// set use track outputs flag
		use_track_outs = preferencesMng->isJackTrackOuts();
		((JackDriver *) driver)->setTrackOuts( use_track_outs );
#else
		audioEngine_raiseError( Hydrogen::NO_JACK_SUPPORT );
		driver = new NullDriver( audioEngine_process );
#endif
	}
	else {
		audioEngine_raiseError( Hydrogen::UNKNOWN_DRIVER );
	}

	// initialize the audio driver
	int res = driver->init( preferencesMng->getBufferSize() );
	if (res != 0) {
		audioEngine_raiseError( Hydrogen::ERROR_STARTING_DRIVER );
		hydrogenInstance->errorLog("Error starting audio driver [audioDriver::init()]");
		delete driver;
		use_track_outs = false;

		driver = new NullDriver( audioEngine_process );
		driver->init(0);
	}

	main_buffer_L = driver->getOut_L();
	main_buffer_R = driver->getOut_R();

#ifdef JACK_SUPPORT
	if (use_track_outs) {
		for (uint i = 0; i < MAX_TRACKS; ++i) {
			track_buffers_L[i] = ((JackDriver*)driver)->getTrackOut_L(i);
			track_buffers_R[i] = ((JackDriver*)driver)->getTrackOut_R(i);
		}
	}
#endif


// check if ALSA_SEQ support is enabled
#ifdef USE_ALSA_SEQ

	// check if midi driver is still alive
	if (midiDriver) {
		hydrogenInstance->errorLog( "The MIDI driver is still alive");
	}

	// Create MIDI driver
	midiDriver = new AlsaMidiDriver();
	midiDriver->open();
	midiDriver->setActive(true);
#endif

	res = driver->connect();
	if (res != 0) {
		audioEngine_raiseError( Hydrogen::ERROR_STARTING_DRIVER );
		hydrogenInstance->errorLog("Error starting audio driver [audioDriver::connect()]");

		delete driver;
		use_track_outs = false;
		driver = new NullDriver( audioEngine_process );
		driver->init(0);
		driver->connect();
	}

	// change the current audio engine state
	if (song == NULL) {
		audioEngine_state = PREPARED;
		hydrogenInstance->infoLog("audioEngine_startAudioDrivers: PREPARED");

	}
	else {
		audioEngine_state = READY;
		hydrogenInstance->infoLog("READY");
	}
	// raise stateChanged event
	for (uint i = 0; i < audioEngine_listeners.size(); i++) {
		audioEngine_listeners[i]->stateChanged(audioEngine_state);
	}
	audioEngine_unlock();
}




//----------------------------------------------------------------------------
/**
 * Stop all audio drivers
 */
void audioEngine_stopAudioDrivers() {
	hydrogenInstance->infoLog("Stopping audio drivers");

	// check current state
	if ( (audioEngine_state != PREPARED) && (audioEngine_state != READY) ) {
		hydrogenInstance->errorLog("audioEngine_stopAudioDrivers: Error the audio engine is not in PREPARED or READY state");
		return;
	}

// check if ALSA_SEQ support is enabled
#ifdef USE_ALSA_SEQ
	hydrogenInstance->infoLog("Stopping MIDI driver");
	// delete MIDI driver
	if (midiDriver) {
		midiDriver->close();
		delete midiDriver;
		midiDriver = NULL;
	}
#endif

	// delete audio driver
	if (driver) {
		driver->disconnect();
		delete driver;
		driver = NULL;
	}

	audioEngine_lock();
	// change the current audio engine state
	audioEngine_state = INITIALIZED;
	hydrogenInstance->infoLog("INITIALIZED");
	// raise stateChanged event
	for (uint i = 0; i < audioEngine_listeners.size(); i++) {
		audioEngine_listeners[i]->stateChanged(INITIALIZED);
	}
	audioEngine_unlock();
}




//----------------------------------------------------------------------------
/**
 * Restart all audio and midi drivers
 */
void audioEngine_restartAudioDrivers() {
	audioEngine_stopAudioDrivers();
	audioEngine_startAudioDrivers();
}






//----------------------------------------------------------------------------
//
// Implementation of Hydrogen class
//
//----------------------------------------------------------------------------






//----------------------------------------------------------------------------
/**
 * Constructor
 */
Hydrogen::Hydrogen(EngineListener *mainListener) {
	if (instance) {
		errorLog( "Hydrogen audio engine is already running" );
	}

	hydrogenInstance = this;
	instance = this;
	audioEngine_init();
	addEngineListener( mainListener );
	audioEngine_startAudioDrivers();
}





//----------------------------------------------------------------------------
/**
 * Destructor
 */
Hydrogen::~Hydrogen() {
	removeSong();
	audioEngine_stopAudioDrivers();
	audioEngine_destroy();
	instance = NULL;
}





//----------------------------------------------------------------------------
/**
 * Return the Hydrogen instance
 */
Hydrogen* Hydrogen::getInstance(EngineListener *listener) {
	if (instance == NULL) {
		instance = new Hydrogen( listener );
	}
	return instance;
}




//----------------------------------------------------------------------------
/**
 * Start the internal sequencer
 */
int Hydrogen::start() {
	return audioEngine_start();
}





//----------------------------------------------------------------------------
/**
 * Stop the internal sequencer
 */
void Hydrogen::stop() {
	audioEngine_stop();
}





//----------------------------------------------------------------------------
/**
 *
 */
void Hydrogen::setSong(Song *newSong) {
	audioEngine_setSong(newSong);
}




//----------------------------------------------------------------------------
/**
 *
 */
void Hydrogen::removeSong() {
	audioEngine_removeSong();
}


//----------------------------------------------------------------------------
/**
 *
 */
Song* Hydrogen::getSong() {
	return song;
}





//----------------------------------------------------------------------------
/**
 *
 */
void Hydrogen::noteOn(Note *note) {
	audioEngine_noteOn(note);
}





//----------------------------------------------------------------------------
/**
 *
 */
void Hydrogen::noteOff(Note *note) {
	audioEngine_noteOff(note);
}




//----------------------------------------------------------------------------
/**
 *
 */
float Hydrogen::getMasterPeak_L() {
	return masterPeak_L;
}




//----------------------------------------------------------------------------
/**
 *
 */
float Hydrogen::getMasterPeak_R() {
	return masterPeak_R;
}





//----------------------------------------------------------------------------
/**
 *
 */
unsigned long Hydrogen::getTickPosition() {
	return audioEngine_getTickPosition();
}






//----------------------------------------------------------------------------
/**
 *
 */
Pattern* Hydrogen::getCurrentPattern() {
	return pattern;
}








//----------------------------------------------------------------------------
/**
 * Set the next pattern (Live mode only)
 */
void Hydrogen::setNextPattern( int pos )
{
	audioEngine_lock();

	if ( song->getMode() == LIVE_MODE ) {
		PatternList *patternList = song->getPatternList();
		if ( (pos >= 0) && ( pos < (int)patternList->getSize() ) ) {
			nextPattern = patternList->get(pos);
		}
		else {
			errorLog("setNextPattern: pos not in patternList range");
			cout << "pos = " << pos << endl;
			cout << "patternListSize = " << patternList->getSize() << endl;
		}
	}
	else {
		errorLog("can't set next pattern in song mode");
	}

	audioEngine_unlock();
}




//----------------------------------------------------------------------------
/**
 *
 */
uint Hydrogen::getPatternPos()
{
	return patternPos;
}



//----------------------------------------------------------------------------
/**
 *
 */
uint Hydrogen::getPlayingNotes()
{
	return playingNotes;
}




//----------------------------------------------------------------------------
/**
 *
 */
void Hydrogen::restartDrivers()
{
	audioEngine_restartAudioDrivers();
}




//----------------------------------------------------------------------------
/**
 * Export a song to a wav file, returns the elapsed time in mSec
 */
uint Hydrogen::exportSong(string filename)
{
	if ( getState() == PLAYING )
	{
		stop();
	}

	int oldEngineMode = song->getMode();
	song->setMode( SONG_MODE );

	bool oldLoopEnabled = song->isLoopEnabled();
	song->setLoopEnabled( false );

	PreferencesMng *preferencesMng = PreferencesMng::getInstance();

	// stop all audio drivers
	audioEngine_stopAudioDrivers();

	/*
		FIXME: Questo codice fa davvero schifo....
	*/
	totalFrames = 0;	// reset total frames
	patternPos = -1;
	audioEngine_state = PLAYING;

	driver = new DiskWriterDriver(audioEngine_process);

	( (DiskWriterDriver*)driver )->setFilename(filename);

	int res = driver->init(preferencesMng->getBufferSize());
	if (res != 0) {
		errorLog("Error starting disk writer driver [DiskWriterDriver::init()]");
	}

	main_buffer_L = driver->getOut_L();
	main_buffer_R = driver->getOut_R();

	uint start = currentTime();


	res = driver->connect();
	if (res != 0) {
		errorLog("Error starting disk writer driver [DiskWriterDriver::connect()]");
	}

	driver->disconnect();

	audioEngine_state = INITIALIZED;
	delete driver;
	driver = NULL;

	main_buffer_L = NULL;
	main_buffer_R = NULL;

	uint finish = currentTime();
	uint elapsedTime = finish - start;

	char tmpbuf[200];
	sprintf(tmpbuf, "Exported in %d msec", elapsedTime);
	hydrogenInstance->infoLog( tmpbuf );

	song->setMode( oldEngineMode );
	song->setLoopEnabled( oldLoopEnabled );

	totalFrames = 0;	// reset total frames
	patternPos = -1;
	audioEngine_startAudioDrivers();

	return elapsedTime;
}





/**
 *
 */
uint Hydrogen::getSongNotesQueue() {
	return songNoteQueue.size();
}



/**
 * Used for display audio driver info
 */
GenericDriver* Hydrogen::getAudioDriver() {
	return driver;
}




/**
 * Used for display midi driver info
 */
AlsaMidiDriver* Hydrogen::getMidiDriver() {
// check if ALSA_SEQ support is enabled
#ifdef USE_ALSA_SEQ
	return midiDriver;
#else
	return NULL;
#endif
}




/**
 *
 */
void Hydrogen::setMasterPeak_L(float value) {
	masterPeak_L = value;
}





/**
 *
 */
void Hydrogen::setMasterPeak_R(float value) {
	masterPeak_R = value;
}


/**
 *
 */
int Hydrogen::getState() {
	return audioEngine_state;
}




/**
 *
 */
void Hydrogen::setCurrentPattern(Pattern *newPattern) {
	audioEngine_lock();

	pattern = newPattern;
	// raise patternChanged event
	for (uint i = 0; i < audioEngine_listeners.size(); i++) {
		audioEngine_listeners[i]->patternChanged();
	}

	audioEngine_unlock();
}




/**
 *
 */
void Hydrogen::addEngineListener(EngineListener *newListener) {
	audioEngine_lock();

	if (newListener) {
		audioEngine_listeners.push_back(newListener);
	}

	audioEngine_unlock();
}




/**
 * Lock the audio engine
 */
void Hydrogen::lockEngine() {
	audioEngine_lock();
}




/**
 * Unlock the audio engine
 */
void Hydrogen::unlockEngine() {
	audioEngine_unlock();
}




/**
 *
 */
float Hydrogen::getProcessTime() {
	return processTime;
}



/**
 *
 */
float Hydrogen::getMaxProcessTime() {
	return maxProcessTime;
}



/**
 *
 */
float Hydrogen::getDelayBufferPeak_L() {
	return delayBufferPeak_L;
}



/**
 *
 */
float Hydrogen::getDelayBufferPeak_R() {
	return delayBufferPeak_R;
}



/**
 *
 */
void Hydrogen::setDelayBufferPeak_L( float value ) {
	delayBufferPeak_L = value;
}



/**
 *
 */
void Hydrogen::setDelayBufferPeak_R( float value ) {
	delayBufferPeak_R = value;
}




/**
 *
 */
int Hydrogen::loadDrumkit( DrumkitInfo *drumkitInfo ) {

	audioEngine_lock();
	infoLog( "loadDrumkit" );

	int oldState = audioEngine_state;
	audioEngine_state = PREPARED;

	audioEngine_clearNoteQueue();

	LocalFileMng fileMng;
	string drumkitPath = fileMng.getDrumkitDirectory( drumkitInfo->getName() );

	InstrumentList *songInstrList = song->getInstrumentList();
	InstrumentList *newInstrList = drumkitInfo->getInstrumentList();
	for (uint i = 0; i < newInstrList->getSize(); i++ ) {
		Instrument *newInstr = newInstrList->get( i );

		string filename = drumkitPath + drumkitInfo->getName() + "/" + newInstr->getSample()->getFilename();

		Instrument *oldInstr = songInstrList->get( i );

		// delete old sample
		Sample *oldSample = oldInstr->getSample();
		delete oldSample;

		// insert new sample from newInstrument
		Sample *newSample = Sample::load( filename );
		newSample->setFilename( newInstr->getSample()->getFilename() );	// restore the "relative path" to drumkit
		oldInstr->setSample( newSample );

		// update instrument properties
		oldInstr->setName( newInstr->getName() );
		oldInstr->setPan_L( newInstr->getPan_L() );
		oldInstr->setPan_R( newInstr->getPan_R() );
		oldInstr->setVolume( newInstr->getVolume() );
		oldInstr->setDrumkitName( newInstr->getDrumkitName() );
		oldInstr->setMuted( newInstr->isMuted() );
	}

//	audioEngine_lock();

	audioEngine_state = oldState;

	infoLog( "loadDrumkit end" );
	audioEngine_unlock();

	return 0;	//ok
}



/**
 *
 */
void Hydrogen::raiseError( uint nErrorCode )
{
	audioEngine_raiseError( nErrorCode );
}




/**
 * Used by AlsaMidiDriver
 */
void Hydrogen::raiseMidiActivityEvent()
{
	for (uint i = 0; i < audioEngine_listeners.size(); i++) {
		audioEngine_listeners[i]->midiActivityEvent();
	}
}



/**
 *
 */
unsigned long Hydrogen::getTotalFrames()
{
	return totalFrames;
}




/**
 * Set the position in the song
 */
void Hydrogen::setPatternPos( uint pos )
{
	audioEngine_lock();

	uint nPatterns = song->getPatternSequence()->getSize();
	if ( pos >= nPatterns )
	{
//		warningLog( "patternPos out of bounds" );
		audioEngine_unlock();
		return;
	}


	int totalTick = 0;
	int patternSize;
	Pattern *pattern = NULL;
	for ( uint i = 0; i < pos; i++ )
	{
		pattern = song->getPatternSequence()->get( i );
		patternSize = pattern->getSize();

		totalTick += patternSize;
	}

	audioEngine_seek( totalTick * tickSize );

	audioEngine_unlock();
}



